BYTE_LENGTH = 8

H0 = {3, 6, 8, 11, 12, 15, 17, 19, 21, 23, 24, 26, 28, 30}


def char_to_int(four_bytes: list) -> int:
    """
    Puts four bytes into a single four byte integer type.

    :param four_bytes: the bytes
    :type four_bytes: list

    :return: the int value
    :rtype int
    """
    num = 0x00
    for i in range(4):
        num = (num << 7) + four_bytes[i]
    return int(num)


def get_bits(buffer: list, start_bit: int, slice_len: int):
    """
    Assumes that end_bit is greater than start_bit and that the result is less than 32 bits, length of an unsigned type.

    :param buffer:
    :param start_bit:
    :param slice_len:

    :return:
    """
    # exclude the last bit of the slice
    end_bit = start_bit + slice_len - 1

    start_byte = start_bit >> 3
    end_byte = end_bit >> 3

    buff_copy = buffer.copy()

    buff_len = len(buffer)
    if end_byte >= buff_len:
        # pad with zeros
        buff_copy.extend([0 for _ in range(end_byte - buff_len + 1)])

    bits = []
    for idx in range(start_byte, end_byte + 1):
        num = buff_copy[idx]
        out = [1 if num & (1 << (BYTE_LENGTH - 1 - n)) else 0 for n in range(BYTE_LENGTH)]
        bits.extend(out)

    # update to relative positions in the bits array
    start_bit %= 8
    end_bit = start_bit + slice_len - 1

    bit_slice = bits[start_bit:end_bit + 1]

    result = 0

    bit_slice.reverse()

    for i in range(len(bit_slice)):
        result += (2 ** i) * bit_slice[i]

    return result


def bit_from_huffman_tables(all_huffman_tables):
    """
    calc the bits from the huffman tables, according to the steganography
    :param all_huffman_tables:
    :return: string contains bits
    """

    s = ""
    for x in all_huffman_tables[-1]:
        if x == 0:
            continue
        s += "0" if x in H0 else "1"
    return s



from mp3stego.decoder.FrameHeader import *
from mp3stego.decoder.util import *


class FrameSideInformation:
    """
    The frame side information class, contains all the information of teh side information of a frame in mp3 file.
    Side information contains info relevant to decode main data.
    """

    def __init__(self):
        self.__main_data_begin: int = 0
        self.__scfsi: np.ndarray = np.zeros((2, 4))

        # Side Info for the two granules. Allocate space for two granules and two channels.
        self.__part2_3_length: np.ndarray = np.zeros((2, 2))
        self.__part2_length: np.ndarray = np.zeros((2, 2))
        self.__big_value: np.ndarray = np.zeros((2, 2))
        self.__global_gain: np.ndarray = np.zeros((2, 2))
        self.__scalefac_compress: np.ndarray = np.zeros((2, 2))
        self.__slen1: np.ndarray = np.zeros((2, 2))
        self.__slen2: np.ndarray = np.zeros((2, 2))
        self.__window_switching: np.ndarray = np.zeros((2, 2))
        self.__block_type: np.ndarray = np.zeros((2, 2))
        self.__mixed_block_flag: np.ndarray = np.zeros((2, 2))
        self.__switch_point_l: np.ndarray = np.zeros((2, 2))
        self.__switch_point_s: np.ndarray = np.zeros((2, 2))
        self.__table_select: np.ndarray = np.zeros((2, 2, 3))
        self.__subblock_gain: np.ndarray = np.zeros((2, 2, 3))
        self.__region0_count: np.ndarray = np.zeros((2, 2))
        self.__region1_count: np.ndarray = np.zeros((2, 2))
        self.__preflag: np.ndarray = np.zeros((2, 2))
        self.__scalefac_scale: np.ndarray = np.zeros((2, 2))
        self.__count1table_select: np.ndarray = np.zeros((2, 2))

        self.__scalefac_l: np.ndarray = np.zeros((2, 2, 22))
        self.__scalefac_s: np.ndarray = np.zeros((2, 2, 3, 13))

    def set_side_info(self, buffer: list, header: FrameHeader):
        """
        The side information contains information on how to decode the main_data.

        :param buffer: buffer that contains the mp3 file, from the first byte of the side info.
        :param header: The frame header
        """

        offset = 0

        # Get main data begin pointer from buffer
        self.__main_data_begin = get_bits(buffer, 0, 9)
        offset += 9
        # Skip private bits
        offset += 5 if header.channel_mode == ChannelMode.Mono else 3

        # Scale factor selection info:
        # If scfsi[scfsi_band] == 1, then scale factors for 1st granule are reused in the 2nd granule.
        # Else, each granule has its own scale factors.
        # scfsi_band indicates what group of scaling factors are reused (1-4)
        for ch in range(header.channels):
            for scfsi_band in range(4):
                self.__scfsi[ch][scfsi_band] = get_bits(buffer, offset, 1) != 0
                offset += 1

        for gr in range(2):
            for ch in range(header.channels):
                # Length of scaling factors and main data in bits.
                self.__part2_3_length[gr][ch] = get_bits(buffer, offset, 12)
                offset += 12
                # Number of values is each big_region.
                self.__big_value[gr][ch] = get_bits(buffer, offset, 9)
                offset += 9
                # Quantizer step size.
                self.__global_gain[gr][ch] = get_bits(buffer, offset, 8)
                offset += 8
                # Used to determine the values of slen1 and slen2.
                self.__scalefac_compress[gr][ch] = get_bits(buffer, offset, 4)
                offset += 4
                # Number of bits given to a range of scale factors.
                # - Normal blocks: slen1 0 - 10, slen2 11-20
                # - Short blocks: Short blocks && mixed_block_flag == 1: slen1 0 - 5, slen2 6-11
                # - Short blocks && mixed_block_flag == 0:
                self.__slen1[gr][ch] = slen[int(self.__scalefac_compress[gr][ch])][0]
                self.__slen2[gr][ch] = slen[int(self.__scalefac_compress[gr][ch])][1]
                # If set, a not normal window is being used.
                self.__window_switching[gr][ch] = get_bits(buffer, offset, 1) == 1
                offset += 1

                if self.__window_switching[gr][ch]:
                    # Window type for the granule: 0=reserved, 1=start block, 2=3 short blocks, 3=end block
                    self.__block_type[gr][ch] = get_bits(buffer, offset, 2)
                    offset += 2
                    # Number of scale factor bands before window switching.
                    self.__mixed_block_flag[gr][ch] = get_bits(buffer, offset, 1) == 1
                    offset += 1
                    if self.__mixed_block_flag[gr][ch]:
                        self.__switch_point_l[gr][ch] = 8
                        self.__switch_point_s[gr][ch] = 3

                    # These are set by default if window_switching is on.
                    self.__region0_count[gr][ch] = 8 if self.__block_type[gr][ch] == 2 else 7
                    # No third region
                    self.__region1_count[gr][ch] = 20 - self.__region0_count[gr][ch]

                    for region in range(2):
                        # Huffman table number for a big region
                        self.__table_select[gr][ch][region] = get_bits(buffer, offset, 5)
                        offset += 5
                    for window in range(3):
                        self.__subblock_gain[gr][ch][window] = get_bits(buffer, offset, 3)
                        offset += 3

                else:
                    # Set by default if window_switching not set.
                    self.__block_type[gr][ch] = 0
                    self.__mixed_block_flag[gr][ch] = False

                    for region in range(3):
                        self.__table_select[gr][ch][region] = get_bits(buffer, offset, 5)
                        offset += 5

                    # Number of scale factor bands in the first big value region.
                    self.__region0_count[gr][ch] = get_bits(buffer, offset, 4)
                    offset += 4
                    # Number of scale factor bands in the third big value region.
                    self.__region1_count[gr][ch] = get_bits(buffer, offset, 3)
                    offset += 3
                    # scale factor bands is 12*3 = 36

                # if set, adds values from a table to the scaling factor
                self.__preflag[gr][ch] = get_bits(buffer, offset, 1)
                offset += 1
                # Determines the step size.
                self.__scalefac_scale[gr][ch] = get_bits(buffer, offset, 1)
                offset += 1
                # Table that determines which count1 table is used.
                self.__count1table_select[gr][ch] = get_bits(buffer, offset, 1)
                offset += 1

    @property
    def main_data_begin(self):
        return self.__main_data_begin

    @property
    def scfsi(self):
        return self.__scfsi

    @property
    def part2_3_length(self):
        return self.__part2_3_length

    @property
    def part2_length(self):
        return self.__part2_length

    @property
    def big_value(self):
        return self.__big_value

    @property
    def global_gain(self):
        return self.__global_gain

    @property
    def scalefac_compress(self):
        return self.__scalefac_compress

    @property
    def slen1(self):
        return self.__slen1

    @property
    def slen2(self):
        return self.__slen2

    @property
    def window_switching(self):
        return self.__window_switching

    @property
    def block_type(self):
        return self.__block_type

    @property
    def mixed_block_flag(self):
        return self.__mixed_block_flag

    @property
    def switch_point_l(self):
        return self.__switch_point_l

    @property
    def switch_point_s(self):
        return self.__switch_point_s

    @property
    def table_select(self):
        return self.__table_select

    @property
    def subblock_gain(self):
        return self.__subblock_gain

    @property
    def region0_count(self):
        return self.__region0_count

    @property
    def region1_count(self):
        return self.__region1_count

    @property
    def preflag(self):
        return self.__preflag

    @property
    def scalefac_scale(self):
        return self.__scalefac_scale

    @property
    def count1table_select(self):
        return self.__count1table_select

    @property
    def scalefac_l(self):
        return self.__scalefac_l

    @property
    def scalefac_s(self):
        return self.__scalefac_s



import os
import sys

import bitarray

from mp3stego import Decoder
from mp3stego import Encoder


# TODO add more information prints when quiet is off
# TODO add more comments
# TODO change function and variables names

def str_to_binary_str(string: str) -> str:
    """
    Convert string into string of bits.

    :param string: the string
    :type string: str

    :return: the string of bits.
    :rtype: str
    """
    binary_str = bitarray.bitarray()
    binary_str.frombytes(string.encode('utf-8'))
    binary_str = [str(ch) for ch in binary_str]
    binary_str = "".join(binary_str)
    return binary_str


class Steganography:
    """
    This class is the Façade of the module. It allows the users a simple way to encode and decode mp3 and wav files,
    hide messages in mp3 files, reveal hidden string and clear mp3 files from any string.

    :param quiet: if False, prints information about the processes and the files.
    :type quiet: bool
    """

    def __init__(self, quiet: bool = True):
        self.quiet = quiet
        self.__encoder: Encoder
        self.__decoder: Decoder
        self.__bitrate: int

    def __encode(self, wav_file_path: str, output_file_path: str, bitrate: int = 320, quiet: bool = True,
                 hide: bool = False, massage: str = ""):
        binary_str = ""
        if hide:
            massage = str(len(massage)) + "#" + massage
            binary_str = str_to_binary_str(massage)

        self.__encoder = Encoder(wav_file_path, output_file_path, bitrate=bitrate, hide_str=binary_str)
        return self.__encoder.encode(quiet=quiet)

    def __decode(self, input_file_path: str, wav_file_path: str, quiet: bool = True, reveal: bool = False,
                 txt_file_path: str = ""):
        self.__decoder = Decoder(input_file_path, wav_file_path)
        self.__bitrate = self.__decoder.decode(quiet, reveal=reveal, txt_file_path=txt_file_path)

    def __delete_wav_file(self, quiet: bool = True):
        self.__decoder.delete_wav_file()
        if not quiet:
            print("Wav file has been deleted.")

    @staticmethod
    def __file_existence(file: str):
        if not os.path.exists(file):
            sys.exit('File not found.')

    def __check_for_decoder(self, input_file_path: str, wav_file_path: str = ""):
        self.__file_existence(input_file_path)
        if wav_file_path == '':
            wav_file_path = input_file_path[:-4] + ".wav"
        if input_file_path[-4:] != '.mp3' or wav_file_path[-4:] != '.wav':
            sys.exit("input_file_path must be mp3 file, wav_file_path must be wav file.")
        return wav_file_path

    def __check_for_encoder(self, wav_file_path: str, output_file_path: str):
        self.__file_existence(wav_file_path)
        if output_file_path[-4:] != '.mp3' or wav_file_path[-4:] != '.wav':
            sys.exit("wav_file_path must be wav file, output_file_path must be mp3 file.")

    def encode_wav_to_mp3(self, wav_file_path: str, output_file_path: str, bitrate: int = 320):
        """
        Allow encoding wav file into mp3 file.

        :param wav_file_path: the wav file path.
        :type wav_file_path: str
        :param output_file_path: the output mp3 file desired path.
        :type output_file_path: str
        :param bitrate: the bitrate of the wav file.
        :type bitrate: int
        """
        if not self.quiet:
            print(f"\n##################\nStart Encoding {wav_file_path} to  {output_file_path}.")
        self.__check_for_encoder(wav_file_path, output_file_path)
        self.__encode(wav_file_path, output_file_path, hide=False, bitrate=bitrate, quiet=self.quiet)
        if not self.quiet:
            print(f"\nFinished Encoding.\n##################")

    def decode_mp3_to_wav(self, input_file_path: str, wav_file_path: str = "") -> int:
        """
        Allow decoding mp3 file into wav file.

        :param input_file_path: the input mp3 file path.
        :type input_file_path: str
        :param wav_file_path: the output wav file desired path.
        :type wav_file_path: str

        :return: the bitrate of the mp3 (and wav) file.
        :rtype: int
        """
        if not self.quiet:
            print(f"\n##################\nStart Decoding {input_file_path} to  {wav_file_path}.")
        wav_file_path = self.__check_for_decoder(input_file_path, wav_file_path)
        self.__decode(input_file_path, wav_file_path, reveal=False, quiet=self.quiet)
        if not self.quiet:
            print(f"\nFinished Decoding.\n##################")
        return self.__bitrate

    def reveal_massage(self, input_file_path: str, txt_file_path: str):
        """
        Allow revealing string from mp3 file. The function writes the string into txt file.

        :param input_file_path: the input mp3 file path.
        :type input_file_path: str
        :param txt_file_path: the output txt file desired path.
        :type txt_file_path: str
        """
        if not self.quiet:
            print(f"\n##################\nStart Revealing hidden message in {input_file_path} to  {txt_file_path}.")
        wav_file_path = self.__check_for_decoder(input_file_path, "")
        if txt_file_path[-4:] != '.txt':
            sys.exit("txt_file_path must be txt file.")
        self.__decode(input_file_path, wav_file_path, reveal=True, quiet=self.quiet, txt_file_path=txt_file_path)
        self.__delete_wav_file(quiet=self.quiet)
        if not self.quiet:
            print(f"\nFinished Revealing.\n##################")

    def hide_message(self, input_file_path: str, output_file_path: str, message: str):
        """
        Allow hiding string in mp3 file. The function creates the new mp3 file with the string hidden in it.

        :param input_file_path: the input mp3 file path.
        :type input_file_path: str
        :param output_file_path: the output mp3 desired path.
        :type output_file_path: str
        :param message: the message to hide in the mp3 file.
        :type message: str

        :return True if the mp3 is too short to contain the message.
        :rtype bool
        """
        if not self.quiet:
            print(f"\n##################\nStart Hiding {message} in {output_file_path}.")
        wav_file_path = self.__check_for_decoder(input_file_path, "")
        self.__decode(input_file_path, wav_file_path, reveal=False, quiet=self.quiet)

        self.__check_for_encoder(wav_file_path, output_file_path)
        too_short = self.__encode(wav_file_path, output_file_path, hide=True, bitrate=self.__bitrate, quiet=self.quiet,
                                  massage=message)
        self.__delete_wav_file(quiet=self.quiet)
        if not self.quiet:
            print(f"\nFinished Hiding.\n##################")
        return too_short

    def clear_file(self, input_file_path: str, output_file_path: str):
        """
        Allow clearing mp3 file from hidden string in it. The function creates new mp3 file without the hidden string.

        :param input_file_path: the input mp3 file path.
        :type input_file_path: str
        :param output_file_path: the output mp3 desired path.
        :type output_file_path: str
        """
        if not self.quiet:
            print(f"\n##################\nStart Cleaning {input_file_path} into {output_file_path}.")
        wav_file_path = self.__check_for_decoder(input_file_path, "")
        self.__decode(input_file_path, wav_file_path, reveal=False, quiet=self.quiet)

        self.__check_for_encoder(wav_file_path, output_file_path)
        self.__encode(wav_file_path, output_file_path, hide=False, bitrate=self.__bitrate, quiet=self.quiet)
        self.__delete_wav_file(quiet=self.quiet)
        if not self.quiet:
            print(f"\nFinished Cleaning.\n##################")



import struct
import sys
from typing import BinaryIO

import numpy as np

from mp3stego.encoder import util


class WavReader:
    """
    class for the wav file, contains all the information of a wav file.

    :param file_path: the wav file path
    :type file_path: str
    :param bit_rate: the bitrate of the wav file.
    :type bit_rate: int
    """

    def __init__(self, file_path: str, bit_rate: int = 320):
        self.__file_path: str = file_path
        self.__bitrate: int = bit_rate
        self.__file: BinaryIO = open(self.__file_path, 'rb')

        self.__read_header()

        self.check_bitrate_index()
        self.check_samplerate_index()

    def __read_header(self):
        """
        Reads the header information to check if it is a valid file with PCM audio samples.
        """
        buffer = self.__file.read(128)

        idx = buffer.find(b'RIFF')  # bytes 1 - 4
        if idx == -1:
            sys.exit('Bad WAVE file.')
        idx += 4
        self.__chunk_size = struct.unpack('<I', buffer[idx:idx + 4])[0]  # bytes 5 - 8

        idx = buffer.find(b'WAVE')  # bytes 9 - 12
        if idx == -1:
            sys.exit('Bad WAVE file.')
        idx = buffer.find(b'fmt ')  # bytes 13 - 16 (format chunk marker)
        if idx == -1:
            sys.exit('Bad WAVE file.')

        idx += 4
        sub_chunk1_size = struct.unpack('<I', buffer[idx:idx + 4])[0]  # bytes 17 - 20, size of fmt section
        if sub_chunk1_size != 16:
            sys.exit('Unsupported WAVE file, compression used instead of PCM.')

        idx += 4
        format_type = struct.unpack('<H', buffer[idx:idx + 2])[0]  # bytes 21 - 22
        if format_type != 1:  # 1 for PCM
            sys.exit('Unsupported WAVE file, compression used instead of PCM.')

        idx += 2
        self.__num_of_ch = struct.unpack('<H', buffer[idx:idx + 2])[0]  # bytes 23 - 24
        if self.__num_of_ch > 1:  # Set to stereo mode if wave data is stereo, mono otherwise.
            self.__mpeg_mode = util.MODES["STEREO"]
        else:
            self.__mpeg_mode = util.MODES["MONO"]

        idx += 2
        self.__samplerate = struct.unpack('<I', buffer[idx:idx + 4])[0]  # bytes 25 - 28
        if self.__samplerate not in (32000, 44100, 48000):
            sys.exit('Unsupported sampling frequency.')
        self.__sample_rate_code = {44100: 0b00, 48000: 0b01, 32000: 0b10}.get(self.__samplerate)

        idx += 4
        self.__byte_rate = struct.unpack('<I', buffer[idx:idx + 4])[0]  # bytes 29 - 32
        # ByteRate = (SampleRate * BitsPerSample * Channels) / 8

        idx += 4
        self.__block_align = struct.unpack('<H', buffer[idx:idx + 2])[0]  # bytes 33 - 34
        # BlockAlign = BitsPerSample * Channels / 8

        idx += 2
        self.__bits_per_sample = struct.unpack('<H', buffer[idx:idx + 2])[0]  # bytes 35 - 36
        if not (self.__bits_per_sample in (8, 16, 32)):
            sys.exit('Unsupported WAVE file, samples not int8, int16 or int32 type.')

        idx = buffer.find(b'data')  # bytes 37 - 40
        if idx == -1:
            sys.exit('Bad WAVE file.')

        idx += 4
        sub_chunk2_size = struct.unpack('<I', buffer[idx:idx + 4])[0]  # bytes 41 - 44, size of data section
        self.__num_of_samples = int(sub_chunk2_size * 8 / self.__bits_per_sample / self.__num_of_ch)

        self.__file.seek(idx + 4)

        self.__num_of_slots = 12 * self.__bitrate * 1000 // self.__samplerate
        self.__copyright = 0
        self.__original = 1
        self.__chmode = 0b11 if self.__num_of_ch == 1 else 0b10
        self.__modext = 0b10
        self.__sync_word = 0b11111111111
        self.__mpeg_version = 0b11
        self.__layer = 0b11
        self.__crc = 0b1
        self.__emphasis = 0b00
        self.__pad_bit = 0
        self.__rest = 0

        self.__buffer = np.fromfile(self.__file, 'int16', self.__num_of_samples * self.__num_of_ch * 2)
        self.__buffer_pos = {0: 0, 1: 1}
        self.__file.close()

    def check_bitrate_index(self):
        if util.find_bitrate_index(self.__bitrate, self.__mpeg_version) < 0:
            sys.exit("Unsupported bitrate configuration.")  # error - not a valid bitrate for encoder

    def check_samplerate_index(self):
        if util.find_samplerate_index(self.__samplerate) < 0:
            sys.exit("Unsupported samplerate configuration.")  # error - not a valid samplerate for encoder

    @property
    def mpeg_mode(self):
        return self.__mpeg_mode

    @property
    def bitrate(self):
        return self.__bitrate

    @property
    def emphasis(self):
        return self.__emphasis

    @property
    def copyright(self):
        return self.__copyright

    @property
    def original(self):
        return self.__original

    @property
    def samplerate(self):
        return self.__samplerate

    @property
    def num_of_channels(self):
        return self.__num_of_ch

    @property
    def file_path(self):
        return self.__file_path

    @property
    def num_of_samples(self):
        return self.__num_of_samples

    @property
    def buffer(self):
        return self.__buffer

    def get_buffer_pos(self, ch):
        return self.__buffer_pos[ch]

    def set_buffer_pos(self, ch, offset):
        self.__buffer_pos[ch] += offset



import math
from enum import Enum

from mp3stego.decoder.tables import *

MAX_BYTE_VALUE = 256


class Band:
    def __init__(self):
        self.long_win: list = []
        self.short_win: list = []


class ChannelMode(Enum):
    Stereo = 0
    JointStereo = 1
    DualChannel = 2
    Mono = 3


class Emphasis(Enum):
    NONE = 0
    MS5015 = 1
    Reserved = 2
    CCITJ17 = 3


class FrameHeader:
    """
    The frame header class, contains all the information of a current header of a frame in mp3 file.
    """

    def __init__(self):
        # Declarations
        self.__buffer: list = []
        self.__mpeg_version: float = 0.0
        self.__layer: int = 0
        self.__crc: bool = False
        self.__bit_rate: int = 0
        self.__sampling_rate: int = 0
        self.__padding: bool = False
        self.__channel_mode: ChannelMode = ChannelMode(0)
        self.__channels: int = 0
        self.__mode_extension: list = [0, 0]
        self.__emphasis: Emphasis = Emphasis(0)
        self.__info: list = [False, False, False]
        self.__band_index: Band = Band()
        self.__band_width: Band = Band()

    def init_header_params(self, buffer):
        """
        Unpack the MP3 header.
        :param buffer: buffer that contains the mp3 file bytes, starts with the first bytes of teh header.
        """
        self.__buffer = buffer
        self.__set_mpeg_version()
        self.__set_layer(self.__buffer[1])
        self.__set_crc()
        self.__set_info()
        self.__set_emphasis()
        self.__set_sampling_rate()
        self.__set_tables()
        self.__set_channel_mode()
        self.__set_mode_extension()
        self.__set_padding()
        self.__set_bit_rate()

    def __set_mpeg_version(self):
        """
        Determine the MPEG version.
        """
        if (self.__buffer[1] & 0x10) == 0x10 and (self.__buffer[1] & 0x08) == 0x08:
            self.__mpeg_version = 1
        elif (self.__buffer[1] & 0x10) == 0x10 and (self.__buffer[1] & 0x08) != 0x08:
            self.__mpeg_version = 2
        elif (self.__buffer[1] & 0x10) != 0x10 and (self.__buffer[1] & 0x08) == 0x08:
            self.__mpeg_version = 0
        elif (self.__buffer[1] & 0x10) != 0x10 and (self.__buffer[1] & 0x08) != 0x08:
            self.__mpeg_version = 2.5

    def __set_layer(self, byte):
        """
        Determine layer
        :param byte: the first byte of the header
        """
        byte = (byte << 5) % MAX_BYTE_VALUE
        byte = (byte >> 6) % MAX_BYTE_VALUE
        self.__layer = 4 - byte

    def __set_crc(self):
        """
        Cyclic redundancy check. If set, two bytes after the header information are used up by the CRC.
        """
        self.__crc = self.__buffer[1] & 0x01

    def __set_info(self):
        """
        Additional information (not important)
        """
        self.__info = [bool(self.__buffer[2] & 0x01), bool(self.__buffer[3] & 0x08), bool(self.__buffer[3] & 0x04)]

    def __set_emphasis(self):
        """
        Although rarely used, there is no method for emphasis.
        """
        value = ((self.__buffer[3] << 6) % MAX_BYTE_VALUE >> 6) % MAX_BYTE_VALUE
        self.__emphasis = Emphasis(value)

    def __set_sampling_rate(self):
        """
        Sampling rate
        """
        rates = [[44100, 48000, 32000], [22050, 24000, 16000], [11025, 12000, 8000]]
        ceil_mpeg_version = int(math.floor(self.__mpeg_version))
        if (self.__buffer[2] & 0x08) != 0x08 and (self.__buffer[2] & 0x04) != 0x04:
            self.__sampling_rate = rates[ceil_mpeg_version - 1][0]
        elif (self.__buffer[2] & 0x08) != 0x08 and (self.__buffer[2] & 0x04) == 0x04:
            self.__sampling_rate = rates[ceil_mpeg_version - 1][1]
        elif (self.__buffer[2] & 0x08) == 0x08 and (self.__buffer[2] & 0x04) != 0x04:
            self.__sampling_rate = rates[ceil_mpeg_version - 1][2]

    def __set_tables(self):
        """
        During the decoding process different tables are used depending on the sampling rate.
        :return:
        """
        if self.__sampling_rate == 32000:
            self.__band_index.short_win = np.array(band_index_table.short_32)
            self.__band_width.short_win = np.array(band_width_table.short_32)
            self.__band_index.long_win = np.array(band_index_table.long_32)
            self.__band_width.long_win = np.array(band_width_table.long_32)
        elif self.__sampling_rate == 44100:
            self.__band_index.short_win = np.array(band_index_table.short_44)
            self.__band_width.short_win = np.array(band_width_table.short_44)
            self.__band_index.long_win = np.array(band_index_table.long_44)
            self.__band_width.long_win = np.array(band_width_table.long_44)
        elif self.__sampling_rate == 48000:
            self.__band_index.short_win = np.array(band_index_table.short_48)
            self.__band_width.short_win = np.array(band_width_table.short_48)
            self.__band_index.long_win = np.array(band_index_table.long_48)
            self.__band_width.long_win = np.array(band_width_table.long_48)

    def __set_channel_mode(self):
        """
        0 -> Stereo
        1 -> Joint stereo (this option requires use of mode_extension)
        2 -> Dual channel
        3 -> Single channel
        """
        value = (self.__buffer[3] >> 6) % MAX_BYTE_VALUE
        self.__channel_mode = ChannelMode(value)
        self.__channels = 1 if self.__channel_mode == ChannelMode.Mono else 2

    def __set_mode_extension(self):
        """
        Applies only to joint stereo.
        """
        if self.__layer == 3:
            self.__mode_extension = [self.__buffer[3] & 0x20, self.__buffer[3] & 0x10]

    def __set_padding(self):
        """
        If set, the frame size is 1 byte larger.
        """
        self.__padding = bool(self.__buffer[2] & 0x02)

    def __set_bit_rate(self):
        """
        For variable bit rate (VBR) files, this data has to be gathered constantly.
        """
        if self.__mpeg_version == 1:
            if self.__layer == 1:
                self.__bit_rate = self.__buffer[2] * 32
            elif self.__layer == 2:
                rates = [32, 48, 56, 64, 80, 96, 112, 128, 160, 192, 224, 256, 320, 384]
                self.__bit_rate = rates[(self.__buffer[2] >> 4) % MAX_BYTE_VALUE - 1] * 1000
            elif self.__layer == 3:
                rates = [32, 40, 48, 56, 64, 80, 96, 112, 128, 160, 192, 224, 256, 320]
                self.__bit_rate = rates[(self.__buffer[2] >> 4) % MAX_BYTE_VALUE - 1] * 1000
            else:
                self.__valid = False
        else:
            if self.__layer == 1:
                rates = [32, 40, 48, 56, 64, 80, 96, 112, 128, 160, 192, 224, 256, 320]
                self.__bit_rate = rates[(self.__buffer[2] >> 4) % MAX_BYTE_VALUE - 1] * 1000
            elif self.__layer < 4:
                rates = [8, 16, 24, 32, 40, 48, 56, 64, 80, 96, 112, 128, 144, 160]
                self.__bit_rate = rates[(self.__buffer[2] >> 4) % MAX_BYTE_VALUE - 1] * 1000
            else:
                self.__valid = False

    @property
    def layer(self):
        return self.__layer

    @property
    def bit_rate(self):
        return self.__bit_rate

    @property
    def sampling_rate(self):
        return self.__sampling_rate

    @property
    def mpeg_version(self):
        return self.__mpeg_version

    @property
    def padding(self):
        return self.__padding

    @property
    def channel_mode(self):
        return self.__channel_mode

    @property
    def channels(self):
        return self.__channels

    @property
    def crc(self):
        return self.__crc

    @property
    def band_index(self):
        return self.__band_index

    @property
    def band_width(self):
        return self.__band_width

    @property
    def mode_extension(self):
        return self.__mode_extension



import os
import sys
import time

from mp3stego.decoder.ID3_Parser import ID3
from mp3stego.decoder.MP3_Parser import MP3Parser


class Decoder:
    """
    Class for wrapping the MP3Parser class, for creating wav file from mp3 file.

    :param file_path: the mp3 file path.
    :type file_path: str
    :param output_file_path: the wav output file path.
    :type output_file_path: str
    """

    def __init__(self, file_path: str, output_file_path: str):
        self.__file_path: str = file_path
        self.__output_file_path: str = output_file_path

        if not os.path.exists(self.__file_path):
            sys.exit('File not found.')

        with open(self.__file_path, 'rb') as f:
            self.__hex_data: list = [c for c in f.read()]

        self.__id3_decoder: ID3 = ID3(self.__hex_data)
        if self.__id3_decoder.is_valid:
            offset = self.__id3_decoder.offset
        else:
            offset = 0

        self.__parser: MP3Parser = MP3Parser(self.__hex_data, offset, self.__output_file_path)

    def __parse_metadata(self, id3_parser: ID3):
        with open('METADATA.txt', 'w') as metadata:
            metadata.write(f'METADATA FOR FILE: {self.__file_path}\n')
            metadata.write('################################\n\n\n')
            metadata.write(f'ID3 Version: {id3_parser.version}\n')
            if len(id3_parser.id3_flags) > 0:
                metadata.write('ID3 Flags:\n')
                for flag in id3_parser.id3_flags:
                    metadata.write(f'- {flag}\n')
                metadata.write('\n')

            metadata.write('\nID3 Frames:\n')
            for i, frame in enumerate(id3_parser.id3_frames):
                metadata.write(f'Frame number: {i}\n')
                metadata.write(f'Frame ID: {frame.id}\n')
                metadata.write(f'Content: {frame.content}\n')
                if len(frame.frame_flags) > 0:
                    metadata.write('Frame Flags:\n')
                    for flag in frame.frame_flags:
                        metadata.write(f'- {flag}\n')
                metadata.write('\n')

    def decode(self, quiet: bool = True, reveal: bool = False, txt_file_path: str = "") -> int:
        """
        Decoding the input mp3 file into wav file. Parse also the metadata.

        :param quiet: if False, print some information about the decoding process
        :type quiet: bool
        :param reveal: if True, reveals the hidden string in the mp3 file.
        :type reveal: bool
        :param txt_file_path: if reveal is True, saves the string into this txt file path.
        :type txt_file_path: str

        :return: the bitrate of the mp3 file (also the bitrate of the output wav file)
        :rtype: int
        """
        if not quiet and self.__id3_decoder.is_valid:
            self.__parse_metadata(self.__id3_decoder)

        start = time.time()
        num_of_parsed_frames = self.__parser.parse_file()
        parsing_time = time.time() - start
        if not quiet:
            print('\nParsed', num_of_parsed_frames, 'frames in', parsing_time, 'seconds.')

        self.__parser.write_to_wav()
        if not quiet:
            print(f"Wav file created on {self.__output_file_path}")

        if reveal:
            if txt_file_path[-4:] != '.txt':
                sys.exit("txt_file_path must be txt file.")

            output_str = ''.join(chr(int(''.join(x), 2)) for x in zip(*[iter(self.__parser.output_bits)] * 8))
            message_len_str = ''
            for idx, ch in enumerate(output_str):
                if ch == '#':
                    break
                message_len_str += ch
            try:
                message_len = int(message_len_str)
            except:
                message_len = 0
                message_len_str = ""

            if (len(message_len_str) + 1 + message_len) > len(output_str):
                output_str = output_str[len(message_len_str) + 1:]
            else:
                output_str = output_str[len(message_len_str) + 1: len(message_len_str) + 1 + message_len]
            f = open(txt_file_path, 'wb')
            f.write(bytes(output_str, 'utf-8'))
            f.close()

        return self.__parser.get_bitrate() // 1000

    def delete_wav_file(self):
        """
        Deletes the output wav file
        """
        if os.path.exists(self.__output_file_path):
            os.remove(self.__output_file_path)


if __name__ == '__main__':
    if len(sys.argv) > 2:
        print("Unexpected number of arguments.")
        exit(-1)
    if len(sys.argv) < 2:
        print("No directory specified.")
        exit(-1)
    file_path = sys.argv[1]

    d = Decoder(file_path, file_path[:-3] + 'wav')
    d.decode()



import os
import sys
import time

from mp3stego.encoder.MP3_Encoder import MP3Encoder
from mp3stego.encoder.WAV_Reader import WavReader


class Encoder:
    """
    Class for wrapping the MP3Encoder class, for creating mp3 file from wav file.

    :param file_path: the wav file path.
    :type file_path: str
    :param output_file_path: the mp3 output file path.
    :type output_file_path: str
    :param bitrate: the bitrate of the input wav file
    :type bitrate: int
    :param hide_str: if is not empty, hides the string inside the output mp3 file.
    :type hide_str: str
    """

    def __init__(self, file_path: str, output_file_path: str, bitrate: int = 320, hide_str: str = ''):
        self.__file_path: str = file_path
        self.__output_file_path: str = output_file_path

        if not os.path.exists(self.__file_path):
            sys.exit('File not found.')

        self.__wav_file: WavReader = WavReader(self.__file_path, bitrate)
        self.__hide_str: str = hide_str
        self.__encoder: MP3Encoder = MP3Encoder(self.__wav_file, hide_str=hide_str)

    def encode(self, quiet: bool = True):
        """
        Encoding the input wav file into mp3 file. can also hide string in the file.

        :param quiet: if False, print some information about the decoding process
        :type quiet: bool

        """
        if not quiet:
            self.__encoder.print_info()
        self.__encoder.encode()

        self.__encoder.write_mp3_file(self.__output_file_path)
        too_short = False
        if not quiet:
            if self.__encoder.hide_str_offset < len(self.__hide_str) - 1:
                print("File too short for this message length, your message has been trimmed.")
                too_short = True
            print(f"MP3 file created on {self.__output_file_path}")
        return too_short


if __name__ == "__main__":
    start = time.time()

    if len(sys.argv) > 2:
        sys.exit('Unexpected number of arguments.')
    if len(sys.argv) < 2:
        sys.exit('No directory specified.')
    file_path = sys.argv[1]

    e = Encoder(file_path, file_path)
    e.encode()

    print(f'Execution time: {int(time.time() - start)} seconds')
